# Tests to write:
#
# analysis:
# TODO: write nupack_multiprocessing tests
# TODO: write Sanger tests
#
# reaction:
# TODO: write ReverseTranscription tests
# TODO: write Transcription tests
# TODO: write Translation tests
# TODO: write Restriction tests
#
# seqio:
# TODO: write read_dna tests
# TODO: write read_seq tests
#
# sequence:
# TODO: dna:
# TODO:   write tests for every public DNA method
# TODO:   write tests for RestrictionSite
# TODO: rna:
# TODO:   write tests for every public RNA method
# TODO: peptide:
# TODO:   write tests for every public Peptide method
